{$WARN UNSAFE_TYPE OFF}
{$WARN UNSAFE_CODE OFF}
{$WARN UNSAFE_CAST OFF}
{$WARN UNIT_PLATFORM OFF}
// (c) Aleksandr Sharahov 2017
// Free for any use
// Changes history
// 2017-01-23 breaking changes
// 2017-01-07 first version
unit ShaDictionaries;

interface

uses
  Classes;

type

{$ifndef UNICODE}
  RawByteString = AnsiString;
{$endif}

  TShaIntegerBoxItems = array of integer;

  TShaIntegerBox = class(TObject)
  private
    FItems: TShaIntegerBoxItems;
    FCount: integer;
    FMaxCount: integer;
    FLast: integer;
    FCode: integer;
  public
    constructor Create(Capacity: integer= 0);
    procedure Clear(Capacity: integer= 0);
    procedure SetCapacity(Capacity: integer);
    function Find(Id: integer): integer;
    function Add(Id: integer): integer;
    procedure Remove(Id: integer);
    function GetNext(var AFromIndex: integer): integer; //starting FromIndex=0
    property Count: integer read FCount;
  end;

  TShaObjectDictionaryItem = packed record
    Key: integer;
    Obj: pointer;
    end;
  PShaObjectDictionaryItem = ^TShaObjectDictionaryItem;
  TShaObjectDictionaryItems = array of TShaObjectDictionaryItem;

  TShaObjectDictionary = class(TObject)
  private
    FItems: TShaObjectDictionaryItems;
    FCount: integer;
    FMaxCount: integer;
    FLast: integer;
    FCode: integer;
    FOwns: boolean;
  protected
    property Items: TShaObjectDictionaryItems read FItems;
  public
    constructor Create(Capacity: integer= 0; OwnsObjects: boolean= false);
    destructor Destroy; override;
    procedure Clear(Capacity: integer= 0);
    procedure SetCapacity(Capacity: integer);
    function Find(Id: integer): integer;
    function Add(Id: integer; AObject: pointer): integer;
    procedure Put(Id: integer; AObject: pointer);
    function Extract(Id: integer): pointer;
    procedure Remove(Id: integer);
    function Get(Id: integer): pointer;
    function GetNext(var AFromIndex: integer): pointer; //starting FromIndex=0
    property Objects[Id: integer]: pointer read Get write Put;
    property Count: integer read FCount;
  end;

  TShaStringHashFunc = function(const Name: RawByteString): integer;
  TShaNamedObjectDictionary = class(TShaObjectDictionary)
  private
    FHash: TShaStringHashFunc;
    FNameOfs: integer;
  protected
    function FindHashed(HashCode: integer; const Name: RawByteString): integer;
    function AddHashed(HashCode: integer; const Name: RawByteString; AObject: pointer): integer;
    procedure PutHashed(HashCode: integer; const Name: RawByteString; AObject: pointer);
    function ExtractHashed(HashCode: integer; const Name: RawByteString): pointer;
    procedure RemoveHashed(HashCode: integer; const Name: RawByteString);
    function GetHashed(HashCode: integer; const Name: RawByteString): pointer;
  public
    constructor Create(Capacity: integer; OwnsObjects: boolean; NameOffset: integer; HashFunc: TShaStringHashFunc= nil);
    function Find(const Name: RawByteString): integer;
    function Add(const Name: RawByteString; AObject: pointer): integer;
    procedure Put(const Name: RawByteString; AObject: pointer);
    function Extract(const Name: RawByteString): pointer;
    procedure Remove(const Name: RawByteString);
    function Get(const Name: RawByteString): pointer;
    property Objects[const Name: RawByteString]: pointer read Get write Put;
  end;

function ShaDictionaryHashAsm(const Name: RawByteString): integer;
function ShaDictionaryHash(const Name: RawByteString): integer;
function ShaPerfectHashAsm(const Name: RawByteString): integer;
function ShaPerfectHash(const Name: RawByteString): integer;
function JenkinsHash(const Name: RawByteString): integer;
function RSHash(const Name: RawByteString): integer;

//----------------------------------------------------------------------------------------------------------------------

implementation

type
  TShaIntegerBoxArea = record a: array[0..0] of integer; end;
  PShaIntegerBoxArea = ^TShaIntegerBoxArea;
  TShaObjectDictionaryArea = record a: array[0..0] of TShaObjectDictionaryItem; end;
  PShaObjectDictionaryArea = ^TShaObjectDictionaryArea;

const
  cMinLength = 4;
  cMaxLength = MaxInt shr 1 + 1;

function GetNewLength(Capacity, Count: integer): integer;
begin;
  if Capacity>cMaxLength then Error(reOutOfMemory);
  if Capacity<Count then Capacity:=Count;
  Result:=cMinLength;
  while (Result-Result shr 2<=Capacity) and (Result<cMaxLength) do inc(Result,Result);
  end;

//----------------------------------------------------------------------------------------------------------------------

procedure IniTShaIntegerBoxItems(pLast: PInteger; iLast: integer);
begin;
  while iLast>=0 do begin;
    pLast^:=iLast; //empty
    dec(pLast);
    dec(iLast);
    end;
  end;

procedure TShaIntegerBox.SetCapacity(Capacity: integer);
var
  OldItems: TShaIntegerBoxItems;
  i, j, NewLength, OldLast, iStart: integer;
begin;
  NewLength:=GetNewLength(Capacity, FCount);
  OldLast:=Length(FItems)-1;
  if (NewLength=OldLast+1)                          //same length
  or (NewLength<=OldLast) and (FCount>0) then exit; //or pack non-empty dictionary

  pointer(OldItems):=pointer(FItems); pointer(FItems):=nil;

  SetLength(FItems, NewLength);
  FLast:=NewLength-1;
  if NewLength=cMaxLength then FMaxCount:=FLast else FMaxCount:=NewLength-NewLength shr 2;

  IniTShaIntegerBoxItems(@FItems[FLast], FLast);

  //move OldItems to FItems
  OldLast:=Length(OldItems)-1;
  if OldLast>=0 then begin;
    i:=OldLast;
    while (i>0) and (OldItems[i]<>i) do dec(i); //find empty cell to start moving
    iStart:=i;
    repeat;
      j:=OldItems[i];
      if j<>i then begin; //not empty
        repeat;
          j:=(j+1) and FLast;
          until FItems[j]=j; //empty
        FItems[j]:=OldItems[i];
        end;
      i:=(i+1) and OldLast;
      until i=iStart;
    end;
  end;

constructor TShaIntegerBox.Create(Capacity: integer= 0);
begin;
  inherited Create;
  SetCapacity(Capacity);
  end;

procedure TShaIntegerBox.Clear(Capacity: integer= 0);
begin;
  FCount:=0;
  if FLast+1=GetNewLength(Capacity, 0)
  then IniTShaIntegerBoxItems(@FItems[FLast], FLast)
  else begin;
    SetLength(FItems, 0);
    SetCapacity(Capacity);
    end;
  end;

function TShaIntegerBox.Find(Id: integer): integer;
var
  c, d: integer;
begin;
  Id:=Id * 1443687719;
  Id:=Id shr 16 xor Id;
  Id:=Id shr 8 xor Id;
  Id:=Id * -1866451833;
  Result:=Id;
  FCode:=Id;
  d:=0;
  while true do begin;
    Result:=(Result+1) and FLast;
    c:=FItems[Result];
    inc(d);                                    //inc our distance
    if (Result-c) and FLast<d then break;      //empty or current distance less than our distance
    if c<>Id then continue;
    Result:=not Result;                        //found
    break;
    end;
  Result:=not Result;
  end;

function TShaIntegerBox.Add(Id: integer): integer;
var
  i, j: integer;
begin;
  if FCount>=FMaxCount then SetCapacity(0);
  Result:=not Find(Id);
  if Result>=0 then begin;
    i:=Result;
    if FItems[i]<>i then begin; //shift
      repeat;
        i:=(i+1) and FLast;
        until FItems[i]=i;
      repeat;
        j:=i;
        i:=(i-1) and FLast;
        FItems[j]:=FItems[i];
        until i=Result;
      end;
    FItems[i]:=FCode;
    inc(FCount);
    end;
  end;

procedure TShaIntegerBox.Remove(Id: integer);
var
  i, j: integer;
begin;
  i:=Find(Id);
  if i>=0 then begin;
    repeat; //try backshift
      j:=i;
      i:=(i+1) and FLast;
      if (i-FItems[i]) and Flast<=1 then break; //empty or unshiftable
      FItems[j]:=FItems[i]; //move raw data
      until false;
    FItems[j]:=j; //empty
    dec(FCount);
    end;
  end;

function TShaIntegerBox.GetNext(var AFromIndex: integer): integer;
var
  i, j: integer;
begin;
  i:=AFromIndex;
  j:=FLast;
  Result:=0;
  with TShaIntegerBoxArea(FItems[0]) do while cardinal(i)<=cardinal(j) do begin;
    if a[i]<>i then begin; //found
      Result:=a[i];
      inc(i);
      break;
      end;
    inc(i);
    end;
  AFromIndex:=i;
  Result:=Result * -1262857929; //-1866451833^-1
  Result:=Result shr 8 xor Result;
  Result:=Result * 630043287; //1443687719^-1
  end;

//----------------------------------------------------------------------------------------------------------------------

procedure InitShaObjectDictionaryItems(pLast: PShaObjectDictionaryItem; iLast: integer);
begin;
  while iLast>=0 do begin;
    pLast.Key:=iLast; //empty
    dec(pLast);
    dec(iLast);
    end;
  end;

procedure FinalizeOwnedShaObjectDictionaryItems(pLast: PShaObjectDictionaryItem; iLast: integer);
begin;
  while iLast>=0 do begin;
    if pLast.Key<>iLast then begin;
      pLast.Key:=iLast; //empty
      TObject(pLast.Obj).Free;
      end;
    dec(pLast);
    dec(iLast);
    end;
  end;

procedure TShaObjectDictionary.SetCapacity(Capacity: integer);
var
  OldItems: TShaObjectDictionaryItems;
  i, j, NewLength, OldLast, iStart: integer;
begin;
  NewLength:=GetNewLength(Capacity, FCount);
  OldLast:=Length(FItems)-1;
  if (NewLength=OldLast+1)                          //same length
  or (NewLength<=OldLast) and (FCount>0) then exit; //or pack non-empty dictionary

  pointer(OldItems):=pointer(FItems); pointer(FItems):=nil;

  SetLength(FItems, NewLength);
  FLast:=NewLength-1;
  if NewLength=cMaxLength then FMaxCount:=FLast else FMaxCount:=NewLength-NewLength shr 2;

  InitShaObjectDictionaryItems(@FItems[FLast], FLast);

  //move OldItems to FItems
  OldLast:=Length(OldItems)-1;
  if OldLast>=0 then begin;
    i:=OldLast;
    while (i>0) and (OldItems[i].Key<>i) do dec(i); //find empty cell to start moving
    iStart:=i;
    repeat;
      j:=OldItems[i].Key;
      if j<>i then begin; //not empty
        repeat;
          j:=(j+1) and FLast;
          until FItems[j].Key=j; //empty
        FItems[j]:=OldItems[i];
        end;
      i:=(i+1) and OldLast;
      until i=iStart;
    end;
  end;

constructor TShaObjectDictionary.Create(Capacity: integer= 0; OwnsObjects: boolean= false);
begin;
  inherited Create;
  FOwns:=OwnsObjects;
  SetCapacity(Capacity);
  end;

procedure TShaObjectDictionary.Clear(Capacity: integer= 0);
begin;
  FCount:=0;
  if FOwns then FinalizeOwnedShaObjectDictionaryItems(@FItems[FLast], FLast);
  if FLast+1=GetNewLength(Capacity, 0) then begin;
    if not FOwns then InitShaObjectDictionaryItems(@FItems[FLast], FLast);
    end
  else begin;
    SetLength(FItems, 0);
    SetCapacity(Capacity);
    end;
  end;

destructor TShaObjectDictionary.Destroy;
begin;
  if FOwns then FinalizeOwnedShaObjectDictionaryItems(@FItems[FLast], FLast);
  inherited;
  end;

function TShaObjectDictionary.Find(Id: integer): integer;
var
  c, d: integer;
begin;
  Id:=Id * 1443687719;
  Id:=Id shr 16 xor Id;
  Id:=Id shr 8 xor Id;
  Id:=Id * -1866451833;
  Result:=Id;
  FCode:=Id;
  d:=0;
  while true do begin;
    Result:=(Result+1) and FLast;
    c:=FItems[Result].Key;
    inc(d);                                    //inc our distance
    if (Result-c) and FLast<d then break;      //empty or current distance less than our distance
    if c<>Id then continue;
    Result:=not Result;                        //found
    break;
    end;
  Result:=not Result;
  end;

procedure ShiftItems(p: PShaObjectDictionaryArea; i, last: integer);
var
  j: integer;
begin;
  with p^ do begin;
    j:=i;
    repeat;
      i:=(i+1) and last;
      until a[i].Key=i;
    if i<j then begin;
      while i>0 do begin;
        a[i]:=a[i-1];
        dec(i);
        end;
      i:=last;
      a[0]:=a[i];
      end;
    while i<>j do begin;
      a[i]:=a[i-1];
      dec(i);
      end;
    end;
  end;

function TShaObjectDictionary.Add(Id: integer; AObject: pointer): integer;
begin;
  if FCount>=FMaxCount then SetCapacity(0);
  Result:=not Find(Id);
  if Result>=0 then with FItems[Result] do begin;
    if Key<>Result then ShiftItems(@FItems[0], Result, FLast);
    Key:=FCode;
    Obj:=AObject;
    inc(FCount);
    end;
  end;

procedure TShaObjectDictionary.Put(Id: integer; AObject: pointer);
var
  i: integer;
begin;
  i:=Add(Id, AObject);
  if i<0 then with FItems[not i] do begin;
    if FOwns and (Obj<>AObject) then TObject(Obj).Free;
    Obj:=AObject;
    end;
  end;

function TShaObjectDictionary.Extract(Id: integer): pointer;
var
  i, j: integer;
begin;
  i:=Find(Id);
  if i<0 then Result:=nil
  else with TShaObjectDictionaryArea(FItems[0]) do begin;
    Result:=a[i].Obj;
    repeat; //try backshift
      j:=i;
      i:=(i+1) and FLast;
      if (i-a[i].Key) and Flast<=1 then break; //empty or unshiftable
      a[j]:=a[i]; //move raw data
      until false;
    a[j].Key:=j; //empty
    dec(FCount);
    end;
  end;

procedure TShaObjectDictionary.Remove(Id: integer);
var
  o: TObject;
begin;
  o:=Extract(Id);
  if FOwns then o.Free;
  end;

function TShaObjectDictionary.Get(Id: integer): pointer;
begin;
  Id:=Find(Id);
  if Id>=0 then Result:=FItems[Id].Obj else Result:=nil;
  end;

function TShaObjectDictionary.GetNext(var AFromIndex: integer): pointer;
var
  i, j: integer;
begin;
  i:=AFromIndex;
  j:=FLast;
  Result:=nil;
  with TShaObjectDictionaryArea(FItems[0]) do while cardinal(i)<=cardinal(j) do begin;
    if a[i].Key<>i then begin; //found
      Result:=a[i].Obj;
      inc(i);
      break;
      end;
    inc(i);
    end;
  AFromIndex:=i;
  end;

//----------------------------------------------------------------------------------------------------------------------

constructor TShaNamedObjectDictionary.Create(Capacity: integer; OwnsObjects: boolean;
  NameOffset: integer; HashFunc: TShaStringHashFunc= nil);
begin;
  inherited Create(Capacity, OwnsObjects);
  FNameOfs:=NameOffset;
  if Assigned(HashFunc) then FHash:=HashFunc else FHash:=ShaPerfectHashAsm;
  end;

function TShaNamedObjectDictionary.Find(const Name: RawByteString): integer;
begin;
  Result:=FindHashed(FHash(Name), Name);
  end;

function TShaNamedObjectDictionary.FindHashed(HashCode: integer; const Name: RawByteString): integer;
var
  c, d: integer;
begin;
  Result:=HashCode;
  d:=0;
  while true do begin;
    Result:=(Result+1) and FLast;
    c:=FItems[Result].Key;
    inc(d);                                    //inc our distance
    if (Result-c) and FLast<d then break;      //empty or current distance less than our distance
    if c<>HashCode then continue;
    if RawByteString(pointer(pAnsiChar(FItems[Result].Obj)+FNameOfs)^)<>Name then continue;
    Result:=not Result;                        //found
    break;
    end;
  Result:=not Result;
  end;

function TShaNamedObjectDictionary.AddHashed(HashCode: integer; const Name: RawByteString; AObject: pointer): integer;
begin;
  if FCount>=FMaxCount then SetCapacity(0);
  Result:=not FindHashed(HashCode, Name);
  if Result>=0 then with FItems[Result] do begin;
    if Key<>Result then ShiftItems(@FItems[0], Result, FLast);
    Key:=HashCode;
    Obj:=AObject;
    inc(FCount);
    end;
  end;

function TShaNamedObjectDictionary.Add(const Name: RawByteString; AObject: pointer): integer;
begin;
  Result:=AddHashed(FHash(Name), Name, AObject);
  end;

procedure TShaNamedObjectDictionary.PutHashed(HashCode: integer; const Name: RawByteString; AObject: pointer);
var
  i: integer;
begin;
  i:=AddHashed(HashCode, Name, AObject);
  if i<0 then with FItems[not i] do begin;
    if FOwns and (Obj<>AObject) then TObject(Obj).Free;
    Obj:=AObject;
    end;
  end;

procedure TShaNamedObjectDictionary.Put(const Name: RawByteString; AObject: pointer);
var
  i: integer;
begin;
  i:=AddHashed(FHash(Name), Name, AObject);
  if i<0 then with FItems[not i] do begin;
    if FOwns and (Obj<>AObject) then TObject(Obj).Free;
    Obj:=AObject;
    end;
  end;

function TShaNamedObjectDictionary.GetHashed(HashCode: integer; const Name: RawByteString): pointer;
begin;
  HashCode:=FindHashed(HashCode, Name);
  if HashCode>=0 then Result:=FItems[HashCode].Obj else Result:=nil;
  end;

function TShaNamedObjectDictionary.Get(const Name: RawByteString): pointer;
var
  i: integer;
begin;
  i:=FindHashed(FHash(Name), Name);
  if i>=0 then Result:=FItems[i].Obj else Result:=nil;
  end;

function TShaNamedObjectDictionary.ExtractHashed(HashCode: integer; const Name: RawByteString): pointer;
var
  i, j: integer;
begin;
  i:=FindHashed(HashCode, Name);
  if i<0 then Result:=nil
  else with TShaObjectDictionaryArea(FItems[0]) do begin;
    Result:=a[i].Obj;
    repeat; //try backshift cells
      j:=i;
      i:=(i+1) and FLast;
      if (i-a[i].Key) and Flast<=1 then break; //empty or unshiftable cell
      a[j]:=a[i]; //move raw data
      until false;
    a[j].Key:=j; //empty
    dec(FCount);
    end;
  end;

function TShaNamedObjectDictionary.Extract(const Name: RawByteString): pointer;
begin;
  Result:=ExtractHashed(FHash(Name), Name);
  end;

procedure TShaNamedObjectDictionary.RemoveHashed(HashCode: integer; const Name: RawByteString);
var
  o: TObject;
begin;
  o:=ExtractHashed(HashCode, Name);
  if FOwns then o.Free;
  end;

procedure TShaNamedObjectDictionary.Remove(const Name: RawByteString);
var
  o: TObject;
begin;
  o:=ExtractHashed(FHash(Name), Name);
  if FOwns then o.Free;
  end;

//----------------------------------------------------------------------------------------------------------------------

//Aleksandr Sharahov
function ShaDictionaryHashAsm(const Name: RawByteString): integer;
asm
  push  ebx
  push  edi
  xor   edx, edx               //Result=0
  test  eax, eax
  jz    @done                  //if pointer(Name)=nil

  mov   ecx, [eax-4]
  cmp   ecx, edx
  jle   @done                  //if Length(Name)<=0

  mov   ebx, -1240289303       //coef:=-1240289303
  lea   ecx, [ecx+eax-1]       //last char

@loop:
  movzx edi, [eax]             //ord(Name[i+1])
  add   edx, edi               //Result:=Result + ord(Name[i+1]
  imul  edx, ebx               //Result:=Result * coef
  imul  ebx, 378551            //coef:=coef * 378551
  cmp   eax, ecx
  jae   @done

  movzx edi, [eax+1]           //ord(Name[i+1])
  add   edx, edi               //Result:=Result + ord(Name[i+1]
  imul  edx, ebx               //Result:=Result * coef
  imul  ebx, 378551            //coef:=coef * 378551
  add   eax, 2                 //next i
  cmp   eax, ecx
  jbe   @loop

@done:
  imul  eax, edx, 1443687719
  pop   edi
  pop   ebx
  end;

//Aleksandr Sharahov
function ShaDictionaryHash(const Name: RawByteString): integer;
var
  coef, i: integer;
begin;
  Result:=0;
  coef:=-1240289303;
  for i:=0 to Length(Name)-1 do begin;
    Result:=(Result+ord(Name[i+1])) * coef;
    coef:=coef * 378551;
    end;
  Result:=Result * 1443687719;
  end;

//Aleksandr Sharahov
function ShaPerfectHashAsm(const Name: RawByteString): integer;
asm
  push  edi
  xor   edx, edx               //Result=0
  test  eax, eax
  jz    @done                  //if pointer(Name)=nil

  mov   ecx, [eax-4]
  cmp   ecx, edx
  jle   @done                  //if Length(Name)<=0
  lea   ecx, [eax+ecx-1]       //last char

@loop:
  imul  edx, 1041204193
  movzx edi, [eax]             //ord(Name[i+1])
  lea   edx, [edx+edi+1507220783]
  movzx edi, [eax+1]           //ord(Name[i+1])
  cmp   eax, ecx
  jae   @done

  imul  edx, 1041204193
  lea   edx, [edx+edi+1507220783]
  add   eax, 2
  cmp   eax, ecx
  ja    @done

  imul  edx, 1041204193
  movzx edi, [eax]             //ord(Name[i+1])
  lea   edx, [edx+edi+1507220783]
  movzx edi, [eax+1]           //ord(Name[i+1])
  cmp   eax, ecx
  jae   @done

  imul  edx, 1041204193
  lea   edx, [edx+edi+1507220783]
  add   eax, 2
  cmp   eax, ecx
  jbe    @loop

@done:
  imul  eax, edx, -1866451833
  pop   edi
  end;

//Aleksandr Sharahov
function ShaPerfectHash(const Name: RawByteString): integer;
var
  i, j: integer;
begin;
  Result:=0;
  for i:=0 to Length(Name)-1 do begin;
    j:=ord(Name[i+1]);
    Result:=Result * 1041204193 + (j+1507220783);
    end;
  Result:=Result * -1866451833;
  end;

//Bob Jenkins
function JenkinsHash(const Name: RawByteString): integer;
var
  i: integer;
begin;
  Result:=0;
  for i:=0 to Length(Name)-1 do begin;
    Result:=Result + ord(Name[i+1]);
    Result:=Result + Result shl 10;
    Result:=Result xor Result shr 6;
    end;
  Result:=Result + Result shl 3;
  Result:=Result xor Result shr 11;
  Result:=Result + Result shl 15;
  end;

//Robert Sedgewick
function RSHash(const Name: RawByteString): integer;
var
  coef, i: integer;
begin;
  Result:=0;
  coef:=63689;
  for i:=0 to Length(Name)-1 do begin;
    Result:=Result * coef + ord(Name[i+1]);
    coef:=coef * 378551;
    end;
  end;

end.


